import os
import sys
import mock
import shutil
import tempfile
import unittest
import yaml

mock_charms = mock.MagicMock()
sys.modules['charms.layer'] = mock_charms
from reactive import alertmanager
from charmhelpers.core import unitdata
from charmhelpers.core.templating import render
from charms.reactive import bus

os.environ['JUJU_UNIT_NAME'] = 'alertmanager'
os.environ['CHARM_DIR'] = '..'


class SimpleConfigMock(dict):
    def __init__(self, *arg, **kw):
        super(SimpleConfigMock, self).__init__(*arg, **kw)
        self._changed_dict = {}
        self._changed_default = True

    def changed(self, key):
        return self._changed_dict.get(key, self._changed_default)

    def set_changed(self, changed_dict):
        self._changed_dict.update(changed_dict)


@mock.patch('os.chown')
@mock.patch('os.fchown')
@mock.patch('reactive.alertmanager.hookenv.open_port')
@mock.patch('reactive.alertmanager.hookenv.status_set')
@mock.patch('reactive.alertmanager.host.service_restart')
@mock.patch('reactive.alertmanager.host.service_running')
@mock.patch('reactive.alertmanager.data_changed')
@mock.patch('reactive.alertmanager.validate_config')
@mock.patch('reactive.alertmanager.hookenv.unit_get')
@mock.patch('reactive.alertmanager.hookenv.config')
@mock.patch('reactive.alertmanager.get_peers')
class TestPrometheusContext(unittest.TestCase):
    def setUp(self):
        super(TestPrometheusContext, self).setUp()
        self._init_tempdir_and_filenames()
        self._init_load_config_yml_defaults()
        self._init_override_render_for_nonroot()

    def _init_tempdir_and_filenames(self):
        self.dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.dir)
        self.maxDiff = None
        self.cfg_yml = os.path.join(self.dir, 'test_alertmanager.yml')
        self.cfg_def = os.path.join(self.dir, 'test_etc_default_alertmanager')
        alertmanager.ALERTMANAGER_YML = self.cfg_yml
        alertmanager.ALERTMANAGER_DEF = self.cfg_def
        # ugly hack, to avoid carrying global unitdata state across tests
        os.environ['UNIT_STATE_DB'] = os.path.join(self.dir, '.unit-state.db')
        unitdata._KV = None

    def _init_load_config_yml_defaults(self):
        # create def_config with parsed config.yaml defaults
        with open("../config.yaml") as fh:
            config_yaml = yaml.safe_load(fh)
            self.def_config = SimpleConfigMock({k: v["default"] for k, v in
                                                config_yaml["options"].items()
                                                if v.get("default")})

    def _init_override_render_for_nonroot(self):
        # need to override render() for 644 perms (ie allow overriding target
        # when running as non-root)
        def render_and_chmod(*args, **kwargs):
            render(*args, **kwargs)
            os.chmod(kwargs['target'], 0o664)
        self.mock_render = mock.patch('reactive.alertmanager.render',
                                      side_effect=render_and_chmod)
        self.mock_render.start()
        self.addCleanup(self.mock_render.stop)

    def _test_enable_generic(self,
                             mock_hookenv_config,
                             configs=None, exp_route=None, exp_receiver=None,
                             *args):
        config = self.def_config
        config.update(configs)
        mock_hookenv_config.return_value = config
        alertmanager.write_alertmanager_config_yml()
        alertmanager.write_alertmanager_config_def()
        yaml_content = yaml.safe_load(open(self.cfg_yml))
        act_route = sorted(yaml_content['route']['routes'] or [])
        self.assertListEqual(act_route, [exp_route])
        act_receiver = yaml_content['receivers'][1]
        self.assertDictEqual(act_receiver, exp_receiver)

    def test_enable_pagerduty(self,
                              mock_get_peers,
                              mock_hookenv_config,
                              *args):
        self._test_enable_generic(
            mock_hookenv_config,
            configs={
                'enable_pagerduty': True,
                'pagerduty_service_key': 'FOOBAR',
            },
            exp_route={
                'receiver': 'juju_receiver',
            },
            exp_receiver={
                'name': 'juju_receiver',
                'pagerduty_configs': [{'service_key': 'FOOBAR'}]
            }
        )

    def test_enable_slack(self,
                          mock_get_peers,
                          mock_hookenv_config,
                          *args):
        self._test_enable_generic(
            mock_hookenv_config,
            configs={
                'enable_slack': True,
                'slack_api_url': 'http://foo:1234/',
                'slack_channel': '#general',
            },
            exp_route={
                'receiver': 'juju_receiver',
            },
            exp_receiver={
                'name': 'juju_receiver',
                'slack_configs': [{'api_url': 'http://foo:1234/',
                                   'channel': '#general'}]
            }
        )

    def test_enable_webhook(self,
                            mock_get_peers,
                            mock_hookenv_config,
                            *args):
        self._test_enable_generic(
            mock_hookenv_config,
            configs={
                'enable_webhook': True,
                'webhook_url': 'http://bar:4567/',
            },
            exp_route={
                'receiver': 'juju_receiver',
            },
            exp_receiver={
                'name': 'juju_receiver',
                'webhook_configs': [{'url': 'http://bar:4567/'}]
            }
        )

    def test_other_configs(self,
                           mock_get_peers,
                           mock_hookenv_config,
                           *args):
        config = self.def_config
        fake_port = '9876'
        fake_ext_url = 'http://foo:1234/bar'
        config.update({
            'port': fake_port,
            'external_url': fake_ext_url,
        })
        mock_hookenv_config.return_value = config
        alertmanager.write_alertmanager_config_yml()
        alertmanager.write_alertmanager_config_def()
        # Verify etc/default/alertmanager has -web.external-url added
        with open(self.cfg_def) as fh:
            contents = fh.read()
            self.assertRegexpMatches(
                contents,
                'ARGS=.*-web.external-url {}.*'.format(fake_ext_url))
            self.assertRegexpMatches(
                contents,
                'ARGS=.*-web.listen.address :{}.*'.format(fake_port))

    @mock.patch('reactive.alertmanager.remove_state')
    def test_install_packages_conditionally_called(self,
                                                   mock_remove_state,
                                                   mock_get_peers,
                                                   mock_hookenv_config,
                                                   mock_unit_get,
                                                   mock_validate_config,
                                                   *args):
        config = self.def_config
        mock_hookenv_config.return_value = config
        # No configs changed, expect no alertmanager-do-install
        config.set_changed({'snap_channel': False})
        alertmanager.check_reconfig_alertmanager()
        self.assertFalse(mock.call('alertmanager.installed')
                         in mock_remove_state.call_args_list)
        # Mock some configs changed, expect alertmanager-do-install
        mock_remove_state.reset_mock()
        config.set_changed({'snap_channel': True})
        alertmanager.check_reconfig_alertmanager()
        self.assertTrue(mock.call('alertmanager.installed')
                        in mock_remove_state.call_args_list)

    def test_configure_alertmanager_relation(self,
                                             mock_get_peers,
                                             mock_hookenv_config,
                                             *args):
        mock_hookenv_config.return_value = self.def_config
        rel_mock = mock.Mock()
        # Verify configure_alertmanager_relation calls configure(port)
        alertmanager.configure_alertmanager_relation(rel_mock)
        self.assertListEqual(rel_mock.method_calls,
                             [mock.call.configure(self.def_config['port'])])

    def test_runtime_args(self,
                          *args):
        alertmanager.runtime_args('-foo', '1')
        alertmanager.runtime_args('-bar', '2')
        self.assertListEqual(alertmanager.runtime_args(), ['-bar 2', '-foo 1'])
        alertmanager.runtime_args('-bar', None)
        self.assertListEqual(alertmanager.runtime_args(), ['-foo 1'])

    @mock.patch('reactive.alertmanager.install_packages')
    # Simulate a full run, from zero to final state,
    # verify we end calling service_restarted(...)
    def test_started_after_full_run(self,
                                    mock_install_packages,
                                    mock_get_peers,
                                    mock_hookenv_config,
                                    mock_hookenv_unit_get,
                                    mock_validate_config,
                                    mock_data_changed,
                                    mock_service_running,
                                    mock_service_restarted,
                                    *args):

        # Feed default config
        config = self.def_config
        mock_hookenv_config.return_value = config
        config.set_changed({'snap_channel': False})
        # Force data_changed() always true
        mock_data_changed.return_value = True

        # Simulate a full reactive run - will consume states until settled:
        bus.dispatch()

        # Verify: main YML present, and calls to: install_packages(),
        # service_restarted(...)
        self.assertTrue(os.path.exists(self.cfg_yml))
        mock_install_packages.called_once_with()
        mock_service_running.called_once_with(alertmanager.SVC_NAME)
        mock_service_restarted.called_once_with(alertmanager.SVC_NAME)
